/*
 * Used to run different UI versions of THE, either explicitly or set a default
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
# else
# include "thedefs.h"
#endif

//#define _WIN32_WINNT 0x0501
#if defined(_WIN32) || defined(_WIN64)
# include <windows.h>
# include <io.h>
# include <time.h>
# include <userenv.h>
# include <malloc.h>
# include <shlobj_core.h>
# define setenv(a,b,c) SetEnvironmentVariable(a,b)
#else
# include <unistd.h>
# include <limits.h>
# include <time.h>
# include <utime.h>
# include <errno.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef __APPLE__
# include <mach-o/dyld.h>
#endif

#if defined(__WATCOMC__)
# include <process.h>
# include <alloca.h>
# include <direct.h>
#endif

#ifndef F_OK
# define F_OK 0
#endif

#if defined(_WIN32) || defined(_WIN64)
# define THE_PATH_SEP '\\'
# define THE_DIR_SEP ';'
# define THE_CONFIG_DIR "THE\\"
# define THE_CONFIG_MODE W_OK
#elif defined(__OS2__) || defined(__MSDOS__) || defined(MSDOS)
# define THE_PATH_SEP '\\'
# define THE_DIR_SEP ';'
# define THE_CONFIG_DIR "config\\THE\\"
# define THE_CONFIG_MODE W_OK
#else
# define THE_PATH_SEP '/'
# define THE_DIR_SEP ':'
# define THE_CONFIG_DIR ".config/THE/"
# define THE_CONFIG_MODE W_OK|X_OK
#endif

#define THE_DEFAULT -1
#define THE_GUI     0
#define THE_SDL     1
#define THE_CONSOLE 2

#if defined(__MSDOS__) || defined(__OS2__) || defined(MSDOS)
# include "thever.h"
#endif

#include "ggets.h"

char *the_version = THE_VERSION;
char *the_release = THE_VERSION_DATE;
char *the_copyright = "Copyright 1991-2023 Mark Hessling";

struct the_variant_t
{
   char *variant_name;
   char *variant_desc;
   size_t marker_file_size;
   int check_x;
   int is_console;
   int call_fork;
   int does_binary_exist;
};
typedef struct the_variant_t THE_VARIANT_T;

/*
 * The following table lists all variants of THE executables
 * When using PDCursesMod or ncurses up to 3 variants of each UI is possible:
 * - variant[w[8]]
 *   where:
 *      variant is UI; eg con, sdl2, gui
 *      w; WIDE_CHAR, but NO UTF8, but includes extra modifiers like A_OVERWRITE
 *      w8; WIDE_CHAR and UTF8, UTF8 and WIDE_CHAR support and extra modifiers like A_OVERWRITE
 *          w8 is work in progress until THE supports UTF8
 */
#if defined(_WIN32) || defined(_WIN64)
#define NUM_VARIANTS 8
#define THE_VARIANT_EXT ".exe"
THE_VARIANT_T variants[NUM_VARIANTS] =
{
   { "guiw",  "GUI with WIDE_CHAR support NO UTF8",0,0,0,0,0 },
   { "gui",  "GUI NO WIDE_CHAR or UTF8 support",0,0,0,0,0 },
   { "sdl2w", "SDLv2 with WIDE_CHAR support NO UTF8",0,0,0,0,0 },
   { "sdl2", "SDLv2 NO WIDE_CHAR or UTF8 support",0,0,0,0,0 },
   { "sdl1w", "SDLv1 with WIDE_CHAR support NO UTF8",0,0,0,0,0 },
   { "sdl1", "SDLv1 NO WIDE_CHAR or UTF8 support",0,0,0,0,0 },
   { "conw",  "console with WIDE_CHAR support NO UTF8",0,0,1,0,0 },
   { "con",  "console NO WIDE_CHAR or UTF8 support",0,0,1,0,0 }
};
#elif defined(OS2) || defined(EMX)
#define NUM_VARIANTS 5
#define THE_VARIANT_EXT ".exe"
THE_VARIANT_T variants[NUM_VARIANTS] =
{
   { "sdl2w","SDLv2 with WIDE_CHAR support",0,0,0,1,0 },
   { "sdl2", "SDLv2 NO UTF8 support",0,0,0,1,0 },
   { "sdl1w","SDLv1 with WIDE_CHAR support",0,0,0,1,0 },
   { "sdl1", "SDLv1 NO UTF8 support",0,0,0,1,0 },
   { "con",  "console NO UTF8 support",0,0,1,0,0 }
};
#elif defined(MSDOS)
#define NUM_VARIANTS 2
#define THE_VARIANT_EXT ".exe"
THE_VARIANT_T variants[NUM_VARIANTS] =
{
   { "vgaw",  "VGA graphics with WIDE_CHAR support",0,0,1,0,0 },
   { "con",  "console NO UTF8 support",0,0,1,0,0 }
};
#else
#define NUM_VARIANTS 12
#define THE_VARIANT_EXT ""
THE_VARIANT_T variants[NUM_VARIANTS] =
{
   { "x11w",   "Native X11 with WIDE_CHAR support NO UTF8",0,1,0,1,0        },
   { "x11",    "Native X11 NO WIDE_CHAR or UTF8 support",0,1,0,1,0         },
   { "sdl2w","SDLv2 with WIDE_CHAR support NO UTF8",0,0,0,1,0       },
   { "sdl2", "SDLv2 NO WIDE_CHAR or UTF8 support",0,0,0,1,0        },
   { "sdl1w","SDLv1 X11 with WIDE_CHAR support NO UTF8",0,0,0,1,0       },
   { "sdl1", "SDLv1 NO WIDE_CHAR or UTF8 support",0,0,0,1,0        },
   { "gl", "Open-GL NO WIDE_CHAR or UTF8 support",0,0,0,1,0        },
   { "glw", "Open-GL with WIDE_CHAR support NO UTF8",0,0,0,1,0        },
   { "conw", "ncurses(console) with WIDE_CHAR support NO UTF8",0,0,1,0,0        },
   { "con",  "ncurses(console) NO WIDE_CHAR or UTF8 support",0,0,1,0,0         },
   { "vt",  "Virtual Terminal NO WIDE_CHAR or UTF8 support",0,0,1,0,0         },
   { "vtw",  "Virtual Terminal with WIDE_CHAR support NO UTF8",0,0,1,0,0         }
};
#endif

#ifdef __APPLE__
char *terms[] =
{
   "Terminal.app",
   "iTerm.app"
};
#endif

#ifdef THE_HOMEBREW_BUILD1
/*
 * HomeBrew on all platform installs each package in its own Cellar.
 * The "the" package which contains this binary is installed in $HOMEBREW_CELLAR/the/[ver]/bin
 * The "the-sdl2w" package is installed in $HOMEBREW_CELLAR/the-sdl2w/[ver]/bin
 * There are symbolic links created for "the" and "the-sdl2w" binaries in $HOMEBREW_CELLAR/bin to the
 * files in each Cellar.  When we get thi sbinary's argv[0] it points to the installed directory not the
 * directory containing the symbolic links, so we can't use the normal mechanism to find our variants.
 * Instead we need to break down argv[0] to $HOMEBREW_CELLAR...
 */
char base_prog[PATH_MAX+1];
int brew_check = 0;
#endif

#if !defined(PATH_MAX)
/* if you don't use PATH_MAX (by #including <limits.h>
 * you will get a crash in realpath!
 */
# define PATH_MAX 2048
#endif
char prog[PATH_MAX+1];

char **my_argv, **in_argv;
int my_argc, in_argc;

int newConsole = 0;
int tracing = 0;

#ifdef __APPLE__
char *termProgram = NULL;
#endif

#if defined(_WIN32) || defined(_WIN64)
void StartupConsole( void )
{
   /*
    * the.exe is a GUI process that can be called from a parent
    * process that has a Console (such as running the.exe from an
    * open command prompt), or from Explorer or Run... in which
    * case the parent does not have a Console.
    * Attempt to attach to the Console of the parent process. If that
    * fails, we already have a Console, so use it. Otherwise, create
    * a Console to use.
    * MHES - 22/2/2013 - we don't attempt this anymore. Firstly it doesn't
    * appear to work, and secondly it precludes THE from running on Win2000
   if ( AttachConsole( ATTACH_PARENT_PROCESS ) == 0 )
    */
   {
      /*
       * We failed to attach to the parent's console, so we need
       * to create our own.
       */
      AllocConsole();
      newConsole = 1;
      freopen( "conin$", "r", stdin );
      freopen( "conout$", "w", stdout );
      freopen( "conout$", "w", stderr );
   }
}

void ClosedownConsole( void )
{
   if ( newConsole )
   {
      fprintf( stderr, "\n==> Press any key to close this window. <==" );
      getchar();
      fclose(stdin);
      fclose(stdout);
      fclose(stderr);
      FreeConsole();
   }
}

int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow )
{
   int rc;
   rc = main( __argc, __argv );
   return rc;
}
#endif

#ifdef __APPLE__
int GetTerminal( void )
{
   if ( termProgram == NULL )
      return 0;
   if ( strcmp( termProgram, "iTerm.app" ) == 0 )
      return 1;
   /* default is Terminal.app */
   return 0;
}
#endif

void ShowMessage( char *msg )
{
#ifdef __APPLE__X
   char *buf,*app;
   if ( getenv( "THE_RUNNING_AS_APP" ) )
      app = "The Hessling Editor";
   else
      app = terms[GetTerminal()];
   buf = alloca( strlen(msg) + strlen(app) + 200 );
   sprintf( buf, "osascript -e 'tell app \"%s\" to display alert \"THE\" message \"%s\" as warning buttons {\"Ok\"}' > /dev/null 2>&1", app, msg );
   system( buf );
#else
# if defined(_WIN32) || defined(_WIN64)
   StartupConsole();
# endif
   fprintf( stderr, "%s", msg );
# if defined(_WIN32) || defined(_WIN64)
   ClosedownConsole();
# endif
#endif
}

int CheckX( int i )
{
   if ( getenv( "DISPLAY" ) == NULL && variants[i].check_x == 1 )
      return 1;
   return 0;
}

int CheckAVariant( char *prog, int i )
{
   char tmp[PATH_MAX+500];
#ifdef THE_HOMEBREW_BUILD1
   if ( brew_check )
   {
      sprintf( base_prog, "%s/../../../the-%s/%s/bin/the-%s", prog, variants[i].variant_name, the_version, variants[i].variant_name );
      if ( realpath( base_prog, tmp ) == NULL )
      {
         return -1;
      }
   }
   else
#endif
   {
      sprintf( tmp, "%sthe-%s%s", prog, variants[i].variant_name, THE_VARIANT_EXT );
   }
   if ( tracing ) fprintf( stderr, "%s %d: variantname %s\n", __FILE__,__LINE__, tmp );
   if ( access( tmp, F_OK ) == 0 )
   {
      return i;
   }
   return -1;
}

int CheckVariant( char *prog, char *var )
{
   int i;
   char tmp[PATH_MAX+500];
   for ( i = 0; i < NUM_VARIANTS; i++ )
   {
      if ( strcmp( var, variants[i].variant_name ) == 0 )
      {
         if ( CheckAVariant( prog, i ) != -1 )
         {
            return i;
         }
         else
         {
            sprintf( tmp, "Specified THE variant: %s is not available\n", var );
            ShowMessage( tmp );
            return -1;
         }
      }
   }
   sprintf( tmp, "Unknown THE variant: %s\n", var );
   ShowMessage( tmp );
   return -1;
}

void usage(char *prog, char *home)
{
   int i;
   char tmp[2048];
   char var[1024];

   strcpy(var,"The following variants are available:\n");
   for ( i = 0; i < NUM_VARIANTS; i++ )
   {
      if ( CheckAVariant( prog, i ) != -1 )
      {
         sprintf( tmp, " - %s - %s\n", variants[i].variant_name, variants[i].variant_desc );
         strcat( var, tmp );
      }
   }
   sprintf(tmp,"\nTHE %s %s %s. All rights reserved.\n"
               "THE is distributed under the terms of the GNU General Public \n"
               "License and comes with NO WARRANTY.\nSee the file COPYING for details.\n"
               "\nUsage:\n\nthe [-h] [-Svar|--set-variant=var] [-Rvar|--run-variant=var] [THE arguments] [file [...]]]\n"
               "the -Svar|--set-variant=var\n"
               "the -Evar|--edit-variant=var\n"
               "\nwhere:\n\n"
               "-h                        show this message\n"
               "-Rvar | --run-variant=var override the default mode and run the specified variant of THE\n"
               "-Svar | --set-variant=var set the variant of THE to run if no specific variant is set via -R or --run-variant switch\n"
               "-Evar | --edit-variant=var edit the variant configuration file\n\n%s\n"
               "\nTHE variant configuration files are located in: %s\n",
               the_version, the_release, the_copyright, var, home);
   ShowMessage( tmp );
}

/*
 * GetArgv0 tries to find the fully qualified filename of the current program.
 * It uses some ugly and system specific tricks and it may return NULL if
 * it can't find any useful value.
 * The argument should be argv[0] of main() and it may be returned.
 * This function must not be called from another as the sole one when starting
 * up.
 */
extern char*_LpPgmName;
static const char *GetArgv0(const char *argv0)
{
#if defined(__WATCOMC__) && defined(MSDOS)
//   argv0 = "the.exe";
   /*
    * Watcom C++ exposes the global variable _LpPgName which points to memoery in the
    * running progam with the fully qualified filename of the running process
    */
   argv0 = _LpPgmName;
#endif
   if ( tracing ) fprintf( stderr, "%s %d: In GetArgv0: argv0: %s\n", __FILE__,__LINE__, argv0 );
#if defined(_WIN32)
   char buf[1024];

   if (GetModuleFileName(NULL, buf, sizeof(buf)) != 0)
      return(strdup(buf)); /* never freed up */
#elif defined(__QNX__) && defined(__WATCOMC__)
   char buffer[PATH_MAX];
   char *buf;
   if ( (buf = _cmdname(buffer) ) != NULL )
      return(strdup(buf)); /* never freed up */
#elif defined(OS2_OLD)
   char buf[512];
   PPIB ppib;

# ifdef __EMX__
   if (_osmode == OS2_MODE)
   {
# endif
      if (DosGetInfoBlocks(NULL, &ppib) == 0)
         if (DosQueryModuleName(ppib->pib_hmte, sizeof(buf), buf) == 0)
            return(strdup(buf));
# ifdef __EMX__
   }
# endif
#endif

#ifdef THE_HOMEBREW_BUILD
   {
      char *brew_prefix = NULL;
      char buf[1024];
      brew_prefix = getenv( "HOMEBREW_PREFIX" );
      if ( brew_prefix )
      {
         strcpy( buf, brew_prefix );
         strcat( buf, "/bin/the" );
         return strdup( buf );
      }
      else
      {
         ShowMessage( "The environment variable HOMEBREW_PREFIX not set; THE cannot determine where variants are.\n" );
         exit( 1 );
      }
   }
#elif defined( __APPLE__ )
   {
      /*
       * will work on MacOS X
       */
      char buf[1024];
      uint32_t result=sizeof(buf);
      if ( _NSGetExecutablePath( buf, &result ) == 0 )
      {
         return strdup( buf );
      }
   }
#else
# if defined(HAVE_READLINK)
   {
      /*
       * will work on Linux 2.1+
       */
      char buf[1024];
      int result;
      result = readlink("/proc/self/exe", buf, sizeof( buf ) );
      if ( ( result > 0 ) && ( result < (int) sizeof( buf ) ) && ( buf[0] != '[' ) )
      {
         buf[result] = '\0';
         return strdup( buf );
      }
   }
# endif
# if defined(HAVE_REALPATH)
   char *prog = realpath( argv0, NULL );
   if ( prog )
   {
      return prog;
   }
# endif
   /*
    * As a last resort we test if the argv0 has no directory delimiter. Then loop through
    * the PATH env variable and test if the file exists in that directory. We then assume if
    * the file exists, then that is where it was launched from.
    */
   if ( strchr( argv0, THE_PATH_SEP ) == NULL )
   {
      int argv0_len = strlen(argv0);
      char *env = getenv( "PATH" );
      if ( env )
      {
         char *myenv = strdup( env );
         if ( myenv )
         {
            char *ptr = myenv;
            char *segment;
            int envlen,i,len = strlen( myenv );
            struct stat sb;
            for ( i = 0; i <= len; i++ )
            {
               if ( myenv[i] == THE_DIR_SEP
               ||   myenv[i] == '\0' ) /* end of string */
               {
                  myenv[i] = '\0';
                  envlen = strlen(ptr);
                  segment = malloc( envlen + 2 + argv0_len );
                  if ( segment )
                  {
                     strcpy( segment, ptr );
                     if ( segment[envlen-1] != THE_PATH_SEP )
                     {
                        segment[envlen] = THE_PATH_SEP;
                        segment[envlen+1] = '\0';
                     }
                     strcat( segment, argv0 );
                     if ( stat( segment, &sb) == 0 )
                     {
                        if ( (sb.st_mode & S_IFMT) == S_IFREG)
                        {
                           /*
                            * We should do more here like checking if the file is exectable (and we are the owner)
                            * or if the file is group executable (and we have the same gid), or other bit is executable
                            * but this will do for now
                            */
                           return segment;
                        }
                     }
                  }
                  /* get the next segment */
                  if ( i != len )
                  {
                     ptr = myenv + i + 1;
                  }
               }
            }
            free( myenv );
         }
      }
   }
#endif
   /* No specific code has found the right file name. Maybe, it's coded
    * in argv0. Check it, if it is an absolute path. Be absolutely sure
    * to detect it safely!
    */
   if ( tracing ) fprintf( stderr, "%s %d: argv0: %s\n", __FILE__,__LINE__, argv0 );
   if (argv0 == NULL)
      return(NULL);

   if ( tracing ) fprintf( stderr, "%s %d: argv0: %s\n", __FILE__,__LINE__, argv0 );
   if (argv0[0] == '/') /* unix systems and some others */
      return(argv0);

   if ( tracing ) fprintf( stderr, "%s %d: argv0: %s\n", __FILE__,__LINE__, argv0 );
   if ((argv0[0] == '\\') && (argv0[1] == '\\')) /* MS and OS/2 UNC names */
      return(argv0);

   if ( tracing ) fprintf( stderr, "%s %d: argv0: %s\n", __FILE__,__LINE__, argv0 );
   if (isalpha(argv0[0]) && (argv0[1] == ':') && (argv0[2] == '\\'))
      return(argv0); /* MS and OS/2 drive letter with path */

   if ( tracing ) fprintf( stderr, "%s %d: argv0: %s\n", __FILE__,__LINE__, argv0 );
   return(NULL); /* not a proven argv0 argument */
}

#if defined(__WATCOMC__) && defined(MSDOS)
int GetDOSDrive( char *dir)
{
   int curdrive = _getdrive();
   int i;
   for ( i = 3; i < 26; i++ )
   {
      if ( _chdrive( i ) == 0 )
      {
         dir[0] = 'A' + i - 1;
         dir[1] = ':';
         dir[2] = '\\';
         dir[3] = '\0';
         _chdrive( curdrive );
         return 0;
      }
   }
   strcpy( dir, "C:\\" );
   return 0;
}
#endif
/*
 * Determines where the special marker files .thegui and .theconsole are read/written
 */
char *GetHomeDirectory()
{
   static char user_dir[1025] = "";
#if defined(_WIN32) || defined(_WIN64)

#if 0
   HANDLE hToken = 0;

   /* Use the Windows API to get the user's profile directory */
   if ( OpenProcessToken( GetCurrentProcess(), TOKEN_QUERY, &hToken ) )
   {
      DWORD BufSize = 1024;

      GetUserProfileDirectory(hToken, user_dir, &BufSize);
      CloseHandle(hToken);
      strcat( user_dir, "\\" );
   }
#else
   PWSTR appdata = NULL;
   if ( SHGetKnownFolderPath( &FOLDERID_RoamingAppData, KF_FLAG_CREATE, NULL, &appdata ) == S_OK)
   {
      wcstombs( user_dir, appdata, 1024 );
      strcat( user_dir, "\\" );
   }
#endif

   /* If it fails set it to the root directory */
   if(!user_dir[0])
   {
      strcpy(user_dir, "C:\\");
   }
#elif defined(__WATCOMC__) && defined(MSDOS)
   GetDOSDrive( user_dir );
#else
   strcpy( user_dir, getenv( "HOME" ) );
   strcat( user_dir, "/" );
#endif
   /* home directory is config dir */
   strcat( user_dir, THE_CONFIG_DIR );
   return user_dir;
}

int CreateConfigDir( char *homedir )
{
   char buf[PATH_MAX+1];
   char err[PATH_MAX+1];
   int len;
   int rc = 0;

   if ( access( homedir, THE_CONFIG_MODE ) != 0 )
   {
      /* config dir doesn't exist or not writeable */
      strcpy( buf, homedir );
      len = strlen( homedir );
      /* c:\users\[user]\appdata\the\ /home/mark/.config/THE/ */
      buf[len-5] = '\0';
      /* does config dir exist ? */
      if ( access ( buf, THE_CONFIG_MODE ) != 0 )
      {
         /* ... no, then create config directory now */
#if defined(__WATCOMC__) || defined(WIN32)
         if ( mkdir( buf ) != 0 )
#else
         if ( mkdir( buf, S_IRWXU ) != 0 )
#endif
         {
            sprintf( err, "Failed to create directory: %s\n", buf );
            ShowMessage( err );
            return 1;
         }
      }
      /* config directory exists; what about THE directory ? */
      buf[len-5] = THE_PATH_SEP;
      buf[len-1] = '\0';
      if ( access ( buf, THE_CONFIG_MODE ) != 0 )
      {
#if defined(__WATCOMC__) || defined(WIN32)
         if ( mkdir( buf ) != 0 )
#else
         if ( mkdir( buf, S_IRWXU ) != 0 )
#endif
         {
            sprintf( err, "Failed to create directory: %s\n", buf );
            ShowMessage( err );
            return 1;
         }
      }
      buf[len-1] = THE_PATH_SEP;
   }
   return rc;
}

int CreateVariantFile( char *homedir, char *var, int idx )
{
   char buf[PATH_MAX+1];
   int fd;

   /* does homedir exist? */
   if ( CreateConfigDir( homedir ) )
   {
      return 1;
   }
   sprintf( buf, "%sthe.var", homedir );
   if ( access( buf, F_OK ) != 0 )
   {
      /* file doesn't exists; create it */
#if defined(_WIN32) || defined(_WIN64) || defined( MSDOS )
      fd = open( buf, O_CREAT|O_WRONLY, _S_IWRITE);
#else
      fd = open( buf, O_CREAT|O_NOCTTY|O_WRONLY, S_IRUSR|S_IWUSR ); /* read/write user */
#endif
      close( fd );
   }
   /* file exists; write num characters so size indicates variant index */
#if defined(_WIN32) || defined(_WIN64) || defined( MSDOS )
   fd = open( buf, O_TRUNC|O_WRONLY, _S_IWRITE);
#else
   fd = open( buf, O_TRUNC|O_NOCTTY|O_WRONLY, S_IRUSR|S_IWUSR ); /* read/write user */
#endif
   /* write idx+1 number of chars from string */
   write( fd, "1234567890", idx + 1 );
   close( fd );

   sprintf( buf, "Successfully set default THE binary: the-%s (%s)\n", var, variants[idx].variant_desc );
   ShowMessage( buf );
   return 0;
}

int CheckVariantFile( char *homedir )
{
   char buf[PATH_MAX+1];
   struct stat attr;

   sprintf( buf, "%sthe.var", homedir );
   if ( stat(buf, &attr) == 0 )
   {
      /* variant index is file size -1 */
      return attr.st_size - 1;
   }
   return -1;
}

int ReadMarkerFileAndSetEnvVariables( char *homedir, int i )
{
   char buf[PATH_MAX+1];
   FILE *fp;
   char *line, *value=NULL, *var;
   int j,len;

   sprintf( buf, "%s%s.tcf", homedir, variants[i].variant_name );
   if ( tracing ) fprintf( stderr, "%s %d: buf %s\n", __FILE__,__LINE__, buf );
   if ( access( buf, F_OK ) == 0 )
   {
      fp = fopen( buf, "r" );
      if ( fp )
      {
         while( fggets_univ( &line, fp ) == 0 )
         {
            /* ignore blank lines */
            if ( strlen( line ) > 0 )
            {
               /* ignore comment lines starting with # or * */
               if ( line[0] != '*'
               &&   line[0] != '#' )
               {
                  var = line;
                  len = strlen( line );
                  for ( j = 0; j < len; j++ )
                  {
                     if ( line[j] == '=' )
                     {
                        line[j] = '\0';
                        value = line+j+1;
                        break;
                     }
                  }
                  setenv( var, value, 1 );
               }
            }
            free(line);
         }
         fclose( fp );
      }
   }
   return 0;
}

#ifdef __APPLE1__
int CheckX11Installed( int terminate, char *msg )
{
   char buf[1024];
   char *app;

   if ( access( "/Applications/Utilities/X11.app", F_OK ) == 0 )
      return 1;
   if ( terminate )
   {
      if ( getenv( "THE_RUNNING_AS_APP" ) )
         app = "The Hessling Editor";
      else
         app = "Terminal";
      sprintf( buf, "osascript -e 'tell app \"%s\" to display alert \"X11 is not installed\" message \"%s\" as warning buttons {\"Ok\"}' > /dev/null 2>&1", app, msg );
      system( buf );
      exit( 1 );
   }
   return 0;
}
#endif

#ifdef __APPLE__
int SetMacroPath( char *runtimedir )
{
   char buf[1024];
   int len = strlen( runtimedir );

   if ( getenv( "THE_RUNNING_AS_APP" ) )
   {
      strncpy( buf, runtimedir, len-6 );
      buf[len-6] = '\0';
      strcat( buf, "Extras" );
      setenv( "THE_MACRO_PATH", buf, 1 );
   }
   return 0;
}
#endif

int ExecuteBinary( char **argv, char *home, char *prog, int idx )
{
//   char program[PATH_MAX+1];
   int rc;
   /* we can now read our marker file and set specified env variables */
   ReadMarkerFileAndSetEnvVariables( home, idx );
   if ( tracing ) fprintf( stderr, "%s %d: prog %s idx %d\n", __FILE__,__LINE__, prog, idx );

   /* now run our actual executable if we have found one */
#ifdef HAVE_FORK
   if ( variants[idx].call_fork )
   {

      int pid = fork();
      if (pid < 0)
      {
         /* Could not fork */
         fprintf( stderr, "Unable to create child process\n" );
         exit(1);
      }
      else if (pid > 0)
      {
         /* Child created ok, so exit parent process */
         _exit(0);
      }
      /* child continues */
   }
#endif
   if ( variants[idx].is_console )
   {
#ifdef __APPLE__
      char prog1[PATH_MAX+1];
#endif
//      sprintf(program, "%sthe-%s", prog, variants[idx].variant_name );
//      argv[0] = prog;
#ifdef __APPLE__
      /* if running as a bundled app, we need to startup the via Terminal */
      if ( getenv( "THE_RUNNING_AS_APP" ) )
      {
         strcpy( prog1, "open -n -W -b com.apple.Terminal \"" );
         strcat( prog1, prog );
         strcat( prog1, "\" &" );
         rc = system( prog1 );
         return rc;
      }
#endif
#if defined(_WIN32) || defined(_WIN64)
      StartupConsole();
#endif
   }
#if defined(__WATCOMC__) && defined(MSDOS)
   /* Watcom C for DOS 32bit does not have exec*() functions */
   rc = spawnv( P_WAIT, prog, argv );
#else
   rc = execv( prog, argv );
   /* should not get here */
#endif
   return rc;
}

static void setVariant( char *var, char *prog, char *home )
{
   int len = strlen( var );
   int idx;
   char *tmp;
   if ( len == 0 )
   {
      ShowMessage( "No variant specified with -S or --set-variant= switch\n" );
      exit( 1 );
   }
   tmp = alloca( strlen(var) + 1 );
   strcpy( tmp, var );
   idx = CheckVariant( prog, tmp );
   if ( idx != -1 )
   {
      CreateVariantFile( home, tmp, idx );
   }
}

#if defined(_WIN32) || defined(_WIN64) || defined(__OS2__) || defined(__MSDOS__) || defined(MSDOS)
char *quoteArgWithSpaces( char *the_arg )
{
   char *tmp;
   if ( strchr( the_arg, ' ' ) != NULL )
   {
      /* known memory leak! */
      tmp = malloc( 2 + strlen( the_arg ) );
      if ( tmp )
      {
         sprintf( tmp, "\"%s\"", the_arg );
         return tmp;
      }
      else
         /* multiple args better than nothing */
         return the_arg;
   }
   return the_arg;
}
#endif

int main( int argc, char *argv[] )
{
   const char *argv0 = NULL;
#ifdef THE_HOMEBREW_BUILD1
   char *brew_cellar = NULL;
#endif
   char *home = NULL;
   int i;
   int len;
   int edit_variant_file = 0;
   int rc=0;
   int path_len = 0;
   int ignore_arg1 = 0;
   int var_idx = -1;
   char the_binary[PATH_MAX+500];

#ifdef __EMX__
   _wildcard(&argc,&argv);
#endif
   /* tracing ? */
   if ( getenv( "THE_LAUNCHER_TRACE" ) )
      tracing = 1;
   /* get our home directory */
   home = GetHomeDirectory();
   if ( tracing ) fprintf( stderr, "%s %d: home %s\n", __FILE__,__LINE__, home );
#ifdef __APPLE__
   char *termProgram = getenv("TERM_PROGRAM");
   /* fudge the X11 requirements for sdl* variants */
   for ( i = 0; i < NUM_VARIANTS; i++ )
   {
      if ( strncmp( variants[i].variant_name, "sdl", 3 ) == 0
      &&   termProgram != NULL )
      {
         variants[i].check_x = 0;
      }
   }
#endif
   /* get our current runtime directory */
   argv0 = GetArgv0( argv[0] );
   if ( tracing ) fprintf( stderr, "%s %d: argv0 %s\n", __FILE__,__LINE__, argv0 );
   if ( argv0 == NULL )
   {
      ShowMessage( "Unable to determine how this program was called\n" );
      exit(1);
   }
   if ( tracing ) fprintf( stderr, "%s %d: argv0 %s\n", __FILE__,__LINE__, argv0 );
#if defined( HAVE_REALPATH ) && !defined( THE_HOMEBREW_BUILD )
   realpath( argv0, prog );
#else
   strcpy( prog, argv0 );
#endif
   if ( tracing ) fprintf( stderr, "%s %d: prog %s\n", __FILE__,__LINE__, prog );
   /* remove the.exe from the full filename */
   len = strlen( prog );
   for ( i = len; i > -1; i--)
   {
      if ( prog[i] == THE_PATH_SEP )
         break;
   }
   path_len = i + 1;
   prog[path_len] = '\0'; /* this is actually the full path to me, excluding program name */
   if ( tracing ) fprintf( stderr, "%s %d: prog %s\n", __FILE__,__LINE__, prog );
#ifdef THE_HOMEBREW_BUILD1
   /*
    * Are we running from a Brew install?
    */
    brew_cellar = getenv( "HOMEBREW_CELLAR" );
    if ( brew_cellar )
    {
       if ( strncmp( brew_cellar, prog, strlen( brew_cellar ) ) == 0 )
       {
          brew_check = 1;
       }
    }
    else
    {
       /* no HOMEBREW_CELLAR set; we still might be running from HOMEBREW_CELLAR...*/
       if ( strncmp( prog, "/usr/local/Cellar", 17 ) == 0 )
       {
          brew_check = 1;
       }
       else
       {
          if ( strncmp( prog, "/opt/homebrew/Cellar", 20 ) == 0 )
          {
             brew_check = 1;
          }
          else
          {
             if ( strncmp( prog, "/home/linuxbrew/.linuxbrew/Cellar", 32 ) == 0 )
             {
                brew_check = 1;
             }
          }
       }
    }
#endif

   /* if displaying version do it and exit; don't run a variant */
   if ( ( argc == 2 && strlen( argv[1] ) == 2 && strcmp( argv[1], "-v" ) == 0 ) )
   {
      printf( "THE %s %s %s. All rights reserved.\n", the_version, the_release, the_copyright );
      exit(0);
   }

   /* if setting a variant, do it and exit */
   if ( argc == 2 && strlen( argv[1] ) >= 14 && strncmp( argv[1], "--set-variant=", 14 ) == 0 )
   {
      setVariant( argv[1]+14, prog, home );
      exit(0);
   }
   if ( argc == 2 && strlen( argv[1] ) >= 2 && strncmp( argv[1], "-S", 2 ) == 0 )
   {
      setVariant( argv[1]+2, prog, home );
      exit(0);
   }
   /* we will now try and run a variant */
   /* see if we are explicitly requesting a variant first */
   if ( argc > 1 )
   {
      char tmp[1024];
      int offset = 0;
      if ( ( strlen( argv[1] ) >= 2 && strncmp( argv[1], "-R", 2 ) == 0 )
        || ( strlen( argv[1] ) >= 14 && strncmp( argv[1], "--run-variant=", 14 ) == 0 ) )
      {
         /* we are going to edit file(s) with a specified variant */
         offset = 14;
      }
      if ( ( strlen( argv[1] ) >= 2 && strncmp( argv[1], "-E", 2 ) == 0 )
        || ( strlen( argv[1] ) >= 14 && strncmp( argv[1], "--edit-variant=", 15 ) == 0 ) )
      {
         if ( argc != 2 )
         {
            ShowMessage( "No filenames allowed when running with -E or --edit-variant switch\n" );
            exit(1);
         }
         /* we are going to edit the variant file */
         offset = 15;
         edit_variant_file = 1;
      }

      if ( offset > 0 )
      {
         if ( strlen( argv[1] ) >= 14 )
         {
            strcpy( tmp, argv[1]+14 );
         }
         else
         {
            strcpy( tmp, argv[1]+2 );
         }
         ignore_arg1 = 1;
         var_idx = CheckVariant( prog, tmp );
         if ( var_idx == -1 )
         {
            /* need to abort here */
            exit( 1 );
         }
         if ( CheckX( var_idx ) == 1 )
         {
            ShowMessage( "The specified variant requires the X11 environment; DISPLAY environment variable is NOT set\n");
            /* need to abort here */
            exit( 1 );
         }
//         CheckMarkerFile( home, var_idx );
      }
   }
   if ( var_idx == -1 )
   {
      /* see if we have a default variant set; .the-variant file exists */
      var_idx = CheckVariantFile( home );
      if ( var_idx != -1 )
      {
         /* we have specified a default, but do we have a binary for it? */
         if ( CheckAVariant( prog, var_idx ) == -1 )
         {
            char tmp[1024];
            sprintf( tmp, "The default variant binary; the-%s is not available\n", variants[var_idx].variant_name);
            ShowMessage( tmp );
            /* need to abort here */
            exit( 1 );
         }
      }
   }
   if ( var_idx == -1 )
   {
      /* find our best variant */
      int xenv_ok = 1;
      for ( i = 0; i < NUM_VARIANTS; i++ )
      {
         xenv_ok = 1;
         if ( CheckAVariant( prog, i ) != -1 )
         {
            variants[i].does_binary_exist = 1;
            if ( CheckX( i ) == 1 )
            {
               xenv_ok = 0;
            }
            if ( variants[i].does_binary_exist == 1 && xenv_ok == 1 )
            {
               var_idx = i;
               break;
            }
         }
      }
   }
   /* if help; show usage and call variant to display it's usage */
   if ( ( argc == 2 && strlen( argv[1] ) == 2 && strcmp( argv[1], "-h" ) == 0 ) )
   {
      usage(prog,home);
   }

   if ( var_idx == -1 )
   {
      ShowMessage( "Cannot find any suitable THE binary to run. Aborting!\n" );
      exit ( 1 );
      /* we have no installed variants; abort */
   }

   /* we have a binary to run; run it */
#ifdef __APPLE__
//   SetMacroPath( prog );
#endif

#ifdef THE_HOMEBREW_BUILD1
   if ( brew_check )
   {
      sprintf( base_prog, "%s/../../../the-%s/%s/bin/the-%s", prog, variants[var_idx].variant_name, the_version, variants[var_idx].variant_name );
      if ( realpath( base_prog, the_binary ) == NULL )
      {
         return -1;
      }
   }
   else
#endif
   {
      /* we have to set the full path to the binary so we can use this in argv[0] for Windows and we can quote it if it has embedded spaces */
      sprintf( the_binary, "%sthe-%s%s", prog, variants[var_idx].variant_name, THE_VARIANT_EXT );
   }
   if ( tracing ) fprintf( stderr, "%s %d: the_binary %s\n", __FILE__,__LINE__, the_binary );

   if ( edit_variant_file == 1 )
   {
      /* edit the variant file only */
      char buf[PATH_MAX+1];
      in_argc = argc+1;
      in_argv = (char **)malloc( sizeof(char *) * (in_argc) );
      in_argv[0] = argv[0];
      in_argv[1] = argv[1];
      sprintf( buf, "%s%s.tcf", home, variants[var_idx].variant_name );
      in_argv[2] = buf;
   }
   else
   {
      in_argc = argc;
      in_argv = (char **)malloc( sizeof(char *) * (in_argc) );
      for ( i = 0; i < argc; i++ )
      {
         in_argv[i] = argv[i];
      }
   }
   my_argv = (char **)malloc( sizeof(char *) * (in_argc+1) );
   my_argc = 0;
   for ( i = 0; i < in_argc; i++ )
   {
      /*
       * We have an arg to pass to the spawned process. If the arg has a space in it, the
       * user MUST have quoted it originally.  Windows (of course) doesn't work like Unix
       * and execv() actually concats all the args into one string to pass to CreateProcess().
       * This results in multiple args for a single arg that has spaces.
       * We have to work around this by appending quotes to args that have spaces for Windows.
       * And crucially argv[0] so we can run from a directory with spaces in it
       */
      if ( !(i == 1 && ignore_arg1) )
      {
#if defined(_WIN32) || defined(_WIN64) || defined(__OS2__) || defined(__MSDOS__) || defined(MSDOS)
         if ( i == 0 )
            my_argv[my_argc++] = quoteArgWithSpaces( the_binary );
         else
            my_argv[my_argc++] = quoteArgWithSpaces( in_argv[i] );
#else
         my_argv[my_argc++] = in_argv[i];
#endif
      }
   }
   /* ensure the argv array is NULL terminated */
   my_argv[my_argc] = NULL;
   /* we can now read our marker file and set specified env variables */
   if ( tracing ) fprintf( stderr, "%s %d: the_binary %s\n", __FILE__,__LINE__, the_binary );
   rc = ExecuteBinary( my_argv, home, the_binary, var_idx );
   free( my_argv );
   free( in_argv );
   exit ( rc );
}
